from gbd_mapping.id import meid

from .healthcare_entity_template import HealthcareEntities, HealthcareEntity

healthcare_entities = HealthcareEntities(
    outpatient_visits=HealthcareEntity(
        name='outpatient_visits',
        kind='healthcare_entity',
        gbd_id=meid(19797),
        utilization=meid(19797),
    ),
    inpatient_visits=HealthcareEntity(
        name='inpatient_visits',
        kind='healthcare_entity',
        gbd_id=meid(18749),
        utilization=meid(18749),
    )
)



from typing import List

import pandas as pd

from vivarium_inputs.globals import gbd


def get_estimation_years(*_, **__) -> pd.Series:
    data = gbd.get_estimation_years()
    return data


def get_year_block(*_, **__) -> pd.DataFrame:
    estimation_years = get_estimation_years()
    year_block = pd.DataFrame({'year_start': range(min(estimation_years), max(estimation_years) + 1)})
    year_block['year_end'] = year_block['year_start'] + 1
    return year_block


def get_age_group_ids(*_, **__) -> List[int]:
    data = gbd.get_age_group_id()
    return data


def get_age_bins(*_, **__) -> pd.DataFrame:
    age_bins = (
        gbd.get_age_bins()[['age_group_id', 'age_group_name', 'age_group_years_start', 'age_group_years_end']]
            .rename(columns={'age_group_years_start': 'age_group_start',
                             'age_group_years_end': 'age_group_end'})
    )
    return age_bins


def get_location_id(location_name):
    return {r.location_name: r.location_id for _, r in gbd.get_location_ids().iterrows()}[location_name]


def get_location_id_parents(location_id: int) -> List[int]:
    location_metadata = gbd.get_location_path_to_global().set_index('location_id')
    parent_ids = [int(loc) for loc in location_metadata.at[location_id, 'path_to_top_parent'].split(',')]
    return parent_ids



from gbd_mapping.base_template import GbdRecord
from gbd_mapping.risk_factor_template import RiskFactor


class AlternativeRiskFactor(RiskFactor):
    """Container for alternative GBD risk factor definitions."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class AlternativeRiskFactors(GbdRecord):
    """Holder of alternative risk factors."""
    __slots__ = ('child_wasting', 'child_underweight', 'child_stunting', )

    def __init__(self,
                 child_stunting: AlternativeRiskFactor,
                 child_underweight: AlternativeRiskFactor,
                 child_wasting: AlternativeRiskFactor,):
        self.child_stunting = child_stunting
        self.child_underweight = child_underweight
        self.child_wasting = child_wasting



from typing import Set

from gbd_mapping import causes, risk_factors, sequelae, covariates, etiologies, coverage_gaps
from vivarium_public_health.dataset_manager import EntityKey

from vivarium_inputs.globals import InvalidQueryError
from vivarium_inputs.interface import (get_measure, get_population_structure, get_age_bins,
                                       get_theoretical_minimum_risk_life_expectancy, get_demographic_dimensions)
from vivarium_inputs.mapping_extension import alternative_risk_factors, healthcare_entities, health_technologies


def loader(entity_key: EntityKey, location: str, modeled_causes: Set[str], all_measures: bool = False):
    entity_data = {
        "cause": {
            "mapping": causes,
            "getter": get_cause_data,
            "measures": ["prevalence", "incidence", "cause_specific_mortality", "disability_weight", "birth_prevalence",
                         "excess_mortality", "remission", "sequelae", "etiologies", "restrictions",],
        },
        "risk_factor": {
            "mapping": risk_factors,
            "getter": get_risk_data,
            "measures": ["affected_causes", "affected_risk_factors", "restrictions", "distribution",
                         "categories", "tmred", "exposure", "exposure_standard_deviation",
                         "relative_risk", "relative_risk_scalar", "population_attributable_fraction",
                         "exposure_distribution_weights"],
        },
        "alternative_risk_factor": {
            "mapping": alternative_risk_factors,
            "getter": get_risk_data,
            "measures": ["affected_causes", "affected_risk_factors", "restrictions", "distribution", "exposure",
                         "exposure_standard_deviation", "exposure_distribution_weights"],
        },
        "sequela": {
            "mapping": sequelae,
            "getter": get_sequela_data,
            "measures": ["healthstate", "prevalence", "incidence", "disability_weight", "birth_prevalence"],
        },
        "healthcare_entity": {
            "mapping": healthcare_entities,
            "getter": get_healthcare_entity_data,
            "measures": ["cost", "utilization"],
         },
        "health_technology": {
             "mapping": health_technologies,
             "getter": get_health_technology_data,
             "measures": ["cost", "effects", "coverage"],
         },
        "coverage_gap": {
            "mapping": coverage_gaps,
            "getter": get_coverage_gap_data,
            "measures": ["affected_causes", "affected_risk_factors", "restrictions", "distribution", "levels",
                         "relative_risk", "exposure"],
        },
        "etiology": {
            "mapping": etiologies,
            "getter": get_etiology_data,
            "measures": ["population_attributable_fraction"],
        },
        "population": {
            "mapping": {'': None},
            "getter": get_population_data,
            "measures": ["structure", "age_bins", "theoretical_minimum_risk_life_expectancy", "demographic_dimensions"],
        },
        "covariate": {
            "mapping": covariates,
            "getter": get_covariate_data,
            "measures": ["estimate"],
        },
    }
    mapping, getter, measures = entity_data[entity_key.type].values()

    entity = mapping[entity_key.name]

    if entity_key.measure not in measures:
        raise InvalidQueryError(f"Unknown measure {entity_key.measure} for entity {entity.name}")

    if not all_measures:
        return getter(entity, entity_key.measure, location, modeled_causes)
    else:
        return data_generator(entity, measures, location, modeled_causes, getter)


def data_generator(entity, measures, location, modeled_causes, getter):
    for measure in measures:
        data = getter(entity, measure, location, modeled_causes)
        if data is not None:
            yield measure, data
            del data


def get_cause_data(cause, measure, location, _):
    if measure in ["sequelae", "etiologies", "restrictions"]:
        data = get_cause_metadata(cause, measure)
    else:
        data = get_measure(cause, measure, location)
    return data


def get_risk_data(risk, measure, location, modeled_causes):
    if measure in ["affected_causes", "affected_risk_factors", "restrictions", "distribution",
                   "categories", "tmred", "relative_risk_scalar"]:
        data = get_risk_metadata(risk, measure, modeled_causes)
    else:
        data = get_measure(risk, measure, location)
    return data


def get_sequela_data(sequela, measure, location, _):
    if measure == "healthstate":
        data = sequela.healthstate.name
    else:
        data = get_measure(sequela, measure, location)
    return data


def get_healthcare_entity_data(healthcare_entity, measure, location, _):
    data = get_measure(healthcare_entity, measure, location)
    return data


def get_health_technology_data(healthcare_technology, measure, location, _):
    if measure in ["effects", "coverage"]:
        raise NotImplementedError()
    data = get_measure(healthcare_technology, measure, location)
    return data


def get_coverage_gap_data(coverage_gap, measure, location, modeled_causes):
    if measure in ["affected_causes", "affected_risk_factors", "restrictions", "distribution", "levels"]:
        data = get_coverage_gap_metadata(coverage_gap, measure, modeled_causes)
    else:
        data = get_measure(coverage_gap, measure, location)
    return data


def get_etiology_data(etiology, measure, location, _):
    data = get_measure(etiology, measure, location)
    return data


def get_population_data(_, measure, location, __):
    if measure == "structure":
        data = get_population_structure(location)
    elif measure == "theoretical_minimum_risk_life_expectancy":
        data = get_theoretical_minimum_risk_life_expectancy()
    elif measure == "age_bins":
        data = get_age_bins()
    else:  # measure == "demographic_dimensions"
        data = get_demographic_dimensions(location)
    return data


def get_covariate_data(covariate, measure, location, _):
    data = get_measure(covariate, measure, location)
    return data


######################
# Metadata unpackers #
######################


def get_cause_metadata(entity, field):
    if field == "restrictions":
        data = entity.restrictions.to_dict()
    else:  # field in ["sequela", "etiologies"]:
        if entity[field]:
            data = [sub_entity.name for sub_entity in entity[field]]
        else:
            data = None
    return data


def get_risk_metadata(risk, measure, modeled_causes):
    if measure in ["restrictions", "categories", "tmred"]:
        if risk[measure] is not None:
            data = risk[measure].to_dict()
        else:
            data = None
    elif measure in ["affected_causes", 'affected_risk_factors']:
        data = [c.name for c in risk.affected_causes if c.name in modeled_causes]
    else:  # measure == "distribution"
        data = risk[measure]
    return data


def get_coverage_gap_metadata(coverage_gap, measure, modeled_causes):
    if measure in ["restrictions", "levels"]:
        data = coverage_gap[measure].to_dict()
    elif measure == "affected_causes":
        data = [c.name for c in coverage_gap.affected_causes if c.name in modeled_causes]
    elif measure == "affected_risk_factors":
        data = [r.name for r in coverage_gap.affected_risk_factors]
    else:  # measure == "distribution"
        data = coverage_gap[measure]
    return data



from typing import Union

from gbd_mapping.base_template import ModelableEntity, GbdRecord
from gbd_mapping.id import meid


class HealthcareEntity(ModelableEntity):
    """Container for healthcare system GBD ids and data."""
    __slots__ = ('name', 'kind', 'gbd_id', 'utilization')

    def __init__(self,
                 name: str,
                 kind: str,
                 gbd_id: Union[meid, None],
                 utilization: meid = None,
                 ):
        super().__init__(name=name, kind=kind, gbd_id=gbd_id)
        self.utilization = utilization


class HealthcareEntities(GbdRecord):
    """Holder of healthcare modelable entities"""
    __slots__ = ('outpatient_visits', 'inpatient_visits')

    def __init__(self,
                 outpatient_visits: HealthcareEntity,
                 inpatient_visits: HealthcareEntity):

        super().__init__()
        self.outpatient_visits = outpatient_visits
        self.inpatient_visits = inpatient_visits



from datetime import datetime
import logging
import warnings
from tables.exceptions import NoSuchNodeError
from typing import Collection, Any
import pandas as pd
from pathlib import Path

from vivarium_public_health.dataset_manager import (EntityKey, Artifact, get_location_term, filter_data)

from vivarium_public_health.disease import DiseaseModel

from vivarium_inputs.data_artifact.loaders import loader
from vivarium_inputs.data_artifact.utilities import get_versions


_log = logging.getLogger(__name__)


class OutdatedArtifactWarning(Warning):
    pass


class ArtifactBuilder:
    configuration_defaults = {
        'input_data': {
            'forecast': False
        }
    }

    def setup(self, builder):
        path = builder.configuration.input_data.artifact_path
        append = builder.configuration.input_data.append_to_artifact
        draw = builder.configuration.input_data.input_draw_number
        self.location = builder.configuration.input_data.location
        self.artifact = self.initialize_artifact(path, append, draw, self.location)

        self.modeled_causes = {c.cause for c in builder.components.get_components(DiseaseModel)}
        self.processed_entities = set()
        self.start_time = datetime.now()

        builder.event.register_listener('post_setup', self.end_processing)

    @staticmethod
    def initialize_artifact(path: str, append: bool, draw: int, location: str) -> Artifact:
        """
        For the given arguments, it checks all the basic conditions and
        initialize the artifact. For now, all the outdated artifacts which
        do not have proper metadata or not consistent metadata will not be
        appended. we will wipe it out and build a new artifact.
        """

        if append:
            if not Path(path).is_file():
                raise ValueError(f'{path} is not a file. You should provide the existing artifact path to append')

            try:
                artifact = Artifact(path, filter_terms=[f'draw == {draw}', get_location_term(location)])

            except NoSuchNodeError:
                #  it means that path was a file but does not have metadata.keyspace inside
                warnings.warn('We will wipe it out and build from scratch', OutdatedArtifactWarning)
                artifact = create_new_artifact(path, draw, location)

            if EntityKey('metadata.locations') not in artifact:
                warnings.warn('We will build from scratch', OutdatedArtifactWarning)
                artifact = create_new_artifact(path, draw, location)

            elif artifact.load('metadata.locations') != [location]:
                raise ValueError(f"Artifact has {artifact.load('metadata.locations')} and we cannot append {location}")

            if EntityKey('metadata.versions') not in artifact:
                warnings.warn('We will build from scratch', OutdatedArtifactWarning)
                artifact = create_new_artifact(path, draw, location)

            elif artifact.load('metadata.versions') != get_versions():
                warnings.warn('Your artifact was made under different versions. We will wipe it out',
                              OutdatedArtifactWarning)
                artifact = create_new_artifact(path, draw, location)
        else:
            artifact = create_new_artifact(path, draw, location)

        return artifact

    def load(self, entity_key: str, future=False, **__) -> Any:
        entity_key = EntityKey(entity_key)
        if entity_key not in self.artifact:
            self.process(entity_key, future)
        data = self.artifact.load(entity_key, future)
        return filter_data(data, **__) if isinstance(data, pd.DataFrame) else data

    def end_processing(self, event) -> None:
        _log.debug(f"Data loading took at most {datetime.now() - self.start_time} seconds")

    def process(self, entity_key: EntityKey, future=False) -> None:
        if entity_key not in self.processed_entities:
            _worker(entity_key, self.location, self.modeled_causes, self.artifact, future)
            self.processed_entities.add(entity_key)


def _worker(entity_key: EntityKey, location: str, modeled_causes: Collection[str],
            artifact: Artifact, future: bool) -> None:
    data = loader(entity_key, location, modeled_causes, all_measures=False)
    artifact.write(entity_key, data)


def create_new_artifact(path: str, draw: int, location: str) -> Artifact:
    if Path(path).is_file():
        Path(path).unlink()
    art = Artifact(path, filter_terms=[f'draw == {draw}', get_location_term(location)])
    art.write('metadata.versions', get_versions())
    art.write('metadata.locations', [location])
    return art



from typing import Union

import pandas as pd

from vivarium_inputs.globals import (gbd, METRICS, MEASURES,
                                     DataAbnormalError, DataDoesNotExistError,
                                     EmptyDataFrameException, NoBestVersionError, InputsException)
from vivarium_inputs.utilities import filter_to_most_detailed_causes
import vivarium_inputs.validation.raw as validation


def extract_data(entity, measure: str, location_id: int) -> Union[pd.Series, pd.DataFrame]:
    extractors = {
        # Cause-like measures
        'incidence': (extract_incidence, {}),
        'prevalence': (extract_prevalence, {}),
        'birth_prevalence': (extract_birth_prevalence, {}),
        'disability_weight': (extract_disability_weight, {}),
        'remission': (extract_remission, {}),
        'deaths': (extract_deaths, {'population': extract_structure}),
        # Risk-like measures
        'exposure': (extract_exposure, {}),
        'exposure_standard_deviation': (extract_exposure_standard_deviation, {'exposure': extract_exposure}),
        'exposure_distribution_weights': (extract_exposure_distribution_weights, {}),
        'relative_risk': (extract_relative_risk, {'exposure': extract_exposure}),
        'population_attributable_fraction': (extract_population_attributable_fraction,
                                             {'exposure': extract_exposure, 'relative_risk': extract_relative_risk}),
        'etiology_population_attributable_fraction': (extract_population_attributable_fraction, {}),
        'mediation_factors': (extract_mediation_factors, {}),
        # Covariate measures
        'estimate': (extract_estimate, {}),
        # Health system measures
        'cost': (extract_cost, {}),
        'utilization': (extract_utilization, {}),
        # Population measures
        'structure': (extract_structure, {}),
        'theoretical_minimum_risk_life_expectancy': (extract_theoretical_minimum_risk_life_expectancy, {}),
    }

    validation.check_metadata(entity, measure)

    try:
        main_extractor, additional_extractors = extractors[measure]
        data = main_extractor(entity, location_id)
        additional_data = {name: extractor(entity, location_id) for name, extractor in additional_extractors.items()}
    except (ValueError, AssertionError, EmptyDataFrameException, NoBestVersionError, InputsException) as e:
        if isinstance(e, ValueError) and f'Metadata associated with rei_id = {entity.gbd_id}' not in str(e):
            raise e
        elif (isinstance(e, AssertionError) and f'Invalid covariate_id {entity.gbd_id}' not in str(e)
                and 'No best model found' not in str(e)):
            raise e
        elif isinstance(e, InputsException) and measure != 'birth_prevalence':
            raise e
        else:
            raise DataDoesNotExistError(f'{measure.capitalize()} data for {entity.name} does not exist.')

    validation.validate_raw_data(data, entity, measure, location_id, **additional_data)
    return data


def extract_prevalence(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_incidence_prevalence(entity_id=entity.gbd_id, location_id=location_id, entity_type=entity.kind)
    data = data[data.measure_id == MEASURES['Prevalence']]
    return data


def extract_incidence(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_incidence_prevalence(entity_id=entity.gbd_id, location_id=location_id, entity_type=entity.kind)
    data = data[data.measure_id == MEASURES['Incidence']]
    return data


def extract_birth_prevalence(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_birth_prevalence(entity_id=entity.gbd_id, location_id=location_id, entity_type=entity.kind)
    data = data[data.measure_id == MEASURES['Incidence']]
    return data


def extract_remission(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_modelable_entity_draws(entity.dismod_id, location_id)
    data = data[data.measure_id == MEASURES['Remission']]
    return data


def extract_disability_weight(entity, location_id: int) -> pd.DataFrame:
    disability_weights = gbd.get_auxiliary_data('disability_weight', entity.kind, 'all', location_id)
    data = disability_weights.loc[disability_weights.healthstate_id == entity.healthstate.gbd_id, :]
    return data


def extract_deaths(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_codcorrect_draws(entity.gbd_id, location_id)
    data = data[data.measure_id == MEASURES['Deaths']]
    return data


def extract_exposure(entity, location_id: int) -> pd.DataFrame:
    if entity.kind == 'risk_factor':
        data = gbd.get_exposure(entity.gbd_id, location_id)
        allowable_measures = [MEASURES['Proportion'], MEASURES['Continuous'], MEASURES['Prevalence']]
        proper_measure_id = set(data.measure_id).intersection(allowable_measures)
        if len(proper_measure_id) != 1:
            raise DataAbnormalError(f'Exposure data have {len(proper_measure_id)} measure id(s). Data should have'
                                    f'exactly one id out of {allowable_measures} but came back with {proper_measure_id}.')
        else:
            data = data[data.measure_id == proper_measure_id.pop()]

    else:  # alternative_risk_factor or coverage_gap
        data = gbd.get_auxiliary_data('exposure', entity.kind, entity.name, location_id)
    return data


def extract_exposure_standard_deviation(entity, location_id: int) -> pd.DataFrame:
    if entity.kind == 'risk_factor':
        data = gbd.get_exposure_standard_deviation(entity.gbd_id, location_id)
    else:  # alternative_risk_factor
        data = gbd.get_auxiliary_data('exposure_standard_deviation', entity.kind, entity.name, location_id)
    return data


def extract_exposure_distribution_weights(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_auxiliary_data('exposure_distribution_weights', entity.kind, entity.name, location_id)
    return data


def extract_relative_risk(entity, location_id: int) -> pd.DataFrame:
    if entity.kind == 'risk_factor':
        data = gbd.get_relative_risk(entity.gbd_id, location_id)
        data = filter_to_most_detailed_causes(data)
    else:  # coverage_gap
        data = gbd.get_auxiliary_data('relative_risk', entity.kind, entity.name, location_id)
    return data


def extract_population_attributable_fraction(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_paf(entity.gbd_id, location_id)
    data = data[data.metric_id == METRICS['Percent']]
    data = data[data.measure_id.isin([MEASURES['YLDs'], MEASURES['YLLs']])]
    data = filter_to_most_detailed_causes(data)
    return data


def extract_mediation_factors(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_auxiliary_data('mediation_factor', entity.kind, entity.name, location_id)
    return data


def extract_estimate(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_covariate_estimate(entity.gbd_id, location_id)
    return data


def extract_cost(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_auxiliary_data('cost', entity.kind, entity.name, location_id)
    return data


def extract_utilization(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_modelable_entity_draws(entity.gbd_id, location_id)
    return data


def extract_structure(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_population(location_id)
    return data


def extract_theoretical_minimum_risk_life_expectancy(entity, location_id: int) -> pd.DataFrame:
    data = gbd.get_theoretical_minimum_risk_life_expectancy()
    return data



from gbd_mapping.base_template import ModelableEntity, GbdRecord


class HealthTechnology(ModelableEntity):
    """Container for health technology ids and data."""
    __slots__ = ('name', 'kind', 'gbd_id')


class HealthTechnologies(GbdRecord):
    """Holder of healthcare technologies."""
    __slots__ = ('hypertension_drugs', )

    def __init__(self,
                 hypertension_drugs: HealthTechnology):
        self.hypertension_drugs = hypertension_drugs



from gbd_mapping.base_template import Restrictions
from gbd_mapping.id import reiid
from gbd_mapping import causes

from .alternative_risk_factor_template import AlternativeRiskFactors, AlternativeRiskFactor

alternative_risk_factors = AlternativeRiskFactors(
    child_underweight=AlternativeRiskFactor(
        name='child_underweight',
        kind='alternative_risk_factor',
        gbd_id=reiid(94),
        level=4,
        most_detailed=True,
        distribution='ensemble',
        population_attributable_fraction_calculation_type='categorical',
        exposure_exists=True,
        exposure_standard_deviation_exists=True,
        exposure_year_type='binned',
        relative_risk_exists=True,
        relative_risk_in_range=False,
        population_attributable_fraction_yll_exists=True,
        population_attributable_fraction_yll_in_range=True,
        population_attributable_fraction_yld_exists=True,
        population_attributable_fraction_yld_in_range=True,
        restrictions=Restrictions(
            male_only=False,
            female_only=False,
            yll_only=False,
            yld_only=False,
            yll_age_group_id_start=4,
            yll_age_group_id_end=235,
            yld_age_group_id_start=2,
            yld_age_group_id_end=235,
            violated=('exposure_age_restriction_violated', 'relative_risk_age_restriction_violated',
                      'population_attributable_fraction_yll_age_restriction_violated', )
        ),
        affected_causes=(causes.all_causes, causes.communicable_maternal_neonatal_and_nutritional_diseases,
                         causes.diarrheal_diseases, causes.lower_respiratory_infections, causes.measles,
                         causes.nutritional_deficiencies, causes.protein_energy_malnutrition,
                         causes.respiratory_infections_and_tuberculosis, causes.enteric_infections,
                         causes.other_infectious_diseases, causes.all_causes,
                         causes.communicable_maternal_neonatal_and_nutritional_diseases, causes.diarrheal_diseases,
                         causes.lower_respiratory_infections, causes.measles, causes.nutritional_deficiencies,
                         causes.protein_energy_malnutrition, causes.respiratory_infections_and_tuberculosis,
                         causes.enteric_infections, causes.other_infectious_diseases, ),
        population_attributable_fraction_of_one_causes=(causes.protein_energy_malnutrition, ),
    ),
    child_wasting=AlternativeRiskFactor(
        name='child_wasting',
        kind='alternative_risk_factor',
        gbd_id=reiid(240),
        level=4,
        most_detailed=True,
        distribution='ensemble',
        population_attributable_fraction_calculation_type='categorical',
        exposure_exists=True,
        exposure_standard_deviation_exists=True,
        exposure_year_type='binned',
        relative_risk_exists=True,
        relative_risk_in_range=False,
        population_attributable_fraction_yll_exists=True,
        population_attributable_fraction_yll_in_range=True,
        population_attributable_fraction_yld_exists=True,
        population_attributable_fraction_yld_in_range=True,
        restrictions=Restrictions(
            male_only=False,
            female_only=False,
            yll_only=False,
            yld_only=False,
            yll_age_group_id_start=4,
            yll_age_group_id_end=235,
            yld_age_group_id_start=2,
            yld_age_group_id_end=235,
            violated=('exposure_age_restriction_violated', 'relative_risk_age_restriction_violated',
                      'population_attributable_fraction_yll_age_restriction_violated', )
        ),
        affected_causes=(causes.all_causes, causes.communicable_maternal_neonatal_and_nutritional_diseases,
                         causes.diarrheal_diseases, causes.lower_respiratory_infections, causes.measles,
                         causes.nutritional_deficiencies, causes.protein_energy_malnutrition,
                         causes.respiratory_infections_and_tuberculosis, causes.enteric_infections,
                         causes.other_infectious_diseases, causes.all_causes,
                         causes.communicable_maternal_neonatal_and_nutritional_diseases, causes.diarrheal_diseases,
                         causes.lower_respiratory_infections, causes.measles, causes.nutritional_deficiencies,
                         causes.protein_energy_malnutrition, causes.respiratory_infections_and_tuberculosis,
                         causes.enteric_infections, causes.other_infectious_diseases,),
        population_attributable_fraction_of_one_causes=(causes.protein_energy_malnutrition,),
    ),
    child_stunting=AlternativeRiskFactor(
        name='child_stunting',
        kind='alternative_risk_factor',
        gbd_id=reiid(241),
        level=4,
        most_detailed=True,
        distribution='ensemble',
        population_attributable_fraction_calculation_type='categorical',
        exposure_exists=True,
        exposure_standard_deviation_exists=None,
        exposure_year_type='binned',
        relative_risk_exists=True,
        relative_risk_in_range=False,
        population_attributable_fraction_yll_exists=True,
        population_attributable_fraction_yll_in_range=True,
        population_attributable_fraction_yld_exists=True,
        population_attributable_fraction_yld_in_range=True,
        restrictions=Restrictions(
            male_only=False,
            female_only=False,
            yll_only=False,
            yld_only=False,
            yll_age_group_id_start=4,
            yll_age_group_id_end=5,
            yld_age_group_id_start=4,
            yld_age_group_id_end=5,
            violated=('exposure_age_restriction_violated', )
        ),
        affected_causes=(causes.all_causes, causes.communicable_maternal_neonatal_and_nutritional_diseases,
                         causes.diarrheal_diseases, causes.lower_respiratory_infections, causes.measles,
                         causes.respiratory_infections_and_tuberculosis, causes.enteric_infections,
                         causes.other_infectious_diseases, causes.all_causes,
                         causes.communicable_maternal_neonatal_and_nutritional_diseases, causes.diarrheal_diseases,
                         causes.lower_respiratory_infections, causes.measles,
                         causes.respiratory_infections_and_tuberculosis, causes.enteric_infections,
                         causes.other_infectious_diseases, ),
        population_attributable_fraction_of_one_causes=(),
    ),
)



"""Global constants, errors, and module imports for inputs processing."""
from gbd_mapping import ModelableEntity, causes, risk_factors


# The purpose of this import block is to mask the dependency on internal
# IHME data and allow CI and automated testing to work.
try:
    from vivarium_gbd_access import gbd
    from gbd_artifacts.exceptions import NoBestVersionError
    from get_draws.api import EmptyDataFrameException, InputsException

except ModuleNotFoundError:
    class GbdDummy:
        """Mock class to wrap internal dependency."""
        def __getattr__(self, item):
            raise ModuleNotFoundError("Required package vivarium_gbd_access not found.")
    gbd = GbdDummy()

    class NoBestVersionError(Exception):
        """Mock class for gbd exception"""
        pass

    class EmptyDataFrameException(Exception):
        """Mock class for gbd exception"""
        pass

    class InputsException(Exception):
        """Mock class for gbd exception"""
        pass


class VivariumInputsError(Exception):
    """Base exception for errors in vivarium_inputs."""
    pass


class DataDoesNotExistError(VivariumInputsError, FileNotFoundError):
    """Exception raised when requested GBD data does not exist."""
    pass


class DataAbnormalError(VivariumInputsError, ValueError):
    """Exception raised when raw GBD data does not have an expected format."""
    pass


class DataTransformationError(VivariumInputsError, ValueError):
    """Exception raised when data is incorrectly transformed."""
    pass


class InvalidQueryError(VivariumInputsError, ValueError):
    """Exception raised when the user makes an invalid query."""
    pass


METRICS = {
    # Mapping of GBD metric names to their GBD ids.
    'Number': 1,
    'Percent': 2,
    'Rate': 3,
    'Rank': 4,
    'Years': 5,
    'p-value': 6,
    'MDG p-value': 7,
    'Probability of death': 8,
    'Index score': 9
}

MEASURES = {
    # Mapping of GBD measure names to their GBD ids.
    'Deaths': 1,
    'DALYs': 2,
    'YLDs': 3,
    'YLLs': 4,
    'Prevalence': 5,
    'Incidence': 6,
    'Remission': 7,
    'Duration': 8,
    'Excess mortality rate': 9,
    'Prevalence * excess mortality rate': 10,
    'Relative risk': 11,
    'Standardized mortality ratio': 12,
    'With-condition mortality rate': 13,
    'All-cause mortality rate': 14,
    'Cause-specific mortality rate': 15,
    'Other cause mortality rate': 16,
    'Case fatality rate': 17,
    'Proportion': 18,
    'Continuous': 19,
    'Survival Rate': 20,
    'Disability Weight': 21,
    'Chronic Prevalence': 22,
    'Acute Prevalence': 23,
    'Acute Incidence': 24,
    'Maternal mortality ratio': 25,
    'Life expectancy': 26,
    'Probability of death': 27,
    'HALE (Healthy life expectancy)': 28,
    'Summary exposure value': 29,
    'Life expectancy no-shock hiv free': 30,
    'Life expectancy no-shock with hiv': 31,
    'Probability of death no-shock hiv free': 32,
    'Probability of death no-shock with hiv': 33,
    'Mortality risk': 34,
    'Short term prevalence': 35,
    'Long term prevalence': 36,
    'Life expectancy decomposition by cause': 37,
    'Birth prevalence': 38,
    'Susceptible population fraction': 39,
    'With Condition population fraction': 40,
    'Susceptible incidence': 41,
    'Total incidence': 42,
    'HAQ Index (Healthcare Access and Quality Index)': 43,
    'Population': 44,
    'Fertility': 45
}

# List of standard demographic id column names.
DEMOGRAPHIC_COLUMNS = ['location_id', 'sex_id', 'age_group_id', 'year_id']
# List of standard GBD draw column names.
DRAW_COLUMNS = [f'draw_{i}' for i in range(1000)]
# Mapping of GBD sex ids
SEXES = {'Male': 1,
         'Female': 2,
         'Combined': 3}
# Mapping of non-standard age group ids sometimes found in GBD data
SPECIAL_AGES = {'all_ages': 22,
                'age_standardized': 27}

# Cause-risk pair where risk may have a protective effect on a certain cause with negative paf
PROTECTIVE_CAUSE_RISK_PAIRS = {
    'high_body_mass_index_in_adults': [causes.neoplasms, causes.breast_cancer, causes.esophageal_cancer]
}

# Keep track of special cases for the sim validator boundary checks where the
# standard boundary in sim validators won't cut it
BOUNDARY_SPECIAL_CASES = {
    'excess_mortality': {
        'Ecuador': {
            'measles': 250_000_000
        }
    }
}

PROBLEMATIC_RISKS = {risk_factors.zinc_deficiency.name:
                         "zinc deficiency relative risk data breaks central comp interpolation."}


class Population(ModelableEntity):
    """Entity wrapper for querying population measures."""
    def __init__(self):
        super().__init__(name='population', kind='population', gbd_id=None)



import pytest

import pandas as pd
import numpy as np

from vivarium_inputs.validation import sim
from vivarium_inputs.globals import DataTransformationError


@pytest.fixture
def mock_validation_context():
    years = pd.DataFrame({'year_start': range(1990, 2017),
                          'year_end': range(1991, 2018)})
    age_bins = pd.DataFrame({'age_group_id': [1, 2, 3, 4, 5],
                             'age_group_name': ['youngest', 'young', 'middle', 'older', 'oldest'],
                             'age_group_start': [0, 1, 15, 45, 60],
                             'age_group_end': [1, 15, 45, 60, 100]})
    context = sim.SimulationValidationContext(
        location='United States',
        years=years,
        age_bins=age_bins
    )

    return context


def test__validate_draw_column_pass():
    df = pd.DataFrame({'draw': range(1000)})
    sim.validate_draw_column(df)


@pytest.mark.parametrize('draws', (900, 1100), ids=('too_few', 'too_many'))
def test_validate_draw_column_incorrect_number(draws):
    df = pd.DataFrame({'draw': range(draws)})
    with pytest.raises(DataTransformationError):
        sim.validate_draw_column(df)


def test_validate_draw_column_missing_column():
    df = pd.DataFrame({'draw_columns': range(1000)})
    with pytest.raises(DataTransformationError, match='in a column named'):
        sim.validate_draw_column(df)


@pytest.mark.parametrize("location", ("Kenya", "Papua New Guinea"))
def test__validate_location_column_pass(mock_validation_context, location):
    mock_validation_context['location'] = location
    df = pd.DataFrame({'location': [location]})
    sim.validate_location_column(df, mock_validation_context)


@ pytest.mark.parametrize('locations,expected_location', (
        (['Kenya', 'Kenya'], 'Egypt'),
        (['Algeria', 'Nigeria'], 'Algeria')
), ids=('mismatch', 'multiple'))
def test_validate_location_column_fail(mock_validation_context, locations, expected_location):
    mock_validation_context['location'] = expected_location
    df = pd.DataFrame({'location': locations})
    with pytest.raises(DataTransformationError):
        sim.validate_location_column(df, mock_validation_context)


def test_validate_location_column_missing_column(mock_validation_context):
    mock_validation_context['location'] = 'Kenya'
    df = pd.DataFrame({'location_column': ['Kenya']})
    with pytest.raises(DataTransformationError, match='in a column named'):
        sim.validate_location_column(df, mock_validation_context)


def test_validate_sex_column_pass():
    df = pd.DataFrame({'sex': ['Male', 'Female']})
    sim.validate_sex_column(df)


@pytest.mark.parametrize("sexes", (
        ('Male', 'female'),
        ('Male', 'Male'),
        ('Female', 'Female')
), ids=('lowercase', 'missing_female', 'missing_male'))
def test_validate_sex_column_fail(sexes):
    df = pd.DataFrame({'sex': sexes})
    with pytest.raises(DataTransformationError):
        sim.validate_sex_column(df)


def test_validate_sex_column_missing_column():
    df = pd.DataFrame({'sex_column': ['Male', 'Female']})
    with pytest.raises(DataTransformationError, match='in a column named'):
        sim.validate_sex_column(df)


def test_validate_age_columns_pass(mock_validation_context):
    ages = mock_validation_context['age_bins'].filter(['age_group_start', 'age_group_end'])
    sim.validate_age_columns(ages, mock_validation_context)


def test_validate_age_columns_invalid_age(mock_validation_context):
    ages = mock_validation_context['age_bins'].filter(['age_group_start', 'age_group_end'])
    ages.loc[2, 'age_group_start'] = -1
    with pytest.raises(DataTransformationError):
        sim.validate_age_columns(ages, mock_validation_context)


def test_validate_age_columns_missing_group(mock_validation_context):
    ages = mock_validation_context['age_bins'].filter(['age_group_start', 'age_group_end'])
    ages.drop(2, inplace=True)
    with pytest.raises(DataTransformationError):
        sim.validate_age_columns(ages, mock_validation_context)


@pytest.mark.parametrize("columns", (('age_group_start',), ('age_group_end',), ('age_group_id_start', 'age_group_end')),
                         ids=('missing_end', 'missing_start', 'typo'))
def test_validate_age_columns_missing_column(columns, mock_validation_context):
    df = pd.DataFrame()
    for col in columns:
        df[col] = [1, 2]
    with pytest.raises(DataTransformationError, match='in columns named'):
        sim.validate_age_columns(df, mock_validation_context)


def test_validate_year_columns_pass(mock_validation_context):
    expected_years = mock_validation_context['years'].sort_values(['year_start', 'year_end'])
    sim.validate_year_columns(expected_years, mock_validation_context)


def test_validate_year_columns_invalid_year(mock_validation_context):
    df = mock_validation_context['years'].sort_values(['year_start', 'year_end'])
    df.loc[2, 'year_end'] = -1
    with pytest.raises(DataTransformationError):
        sim.validate_year_columns(df, mock_validation_context)


def test__validate_year_columns_missing_group(mock_validation_context):
    df = mock_validation_context['years'].sort_values(['year_start', 'year_end'])
    df.drop(0, inplace=True)
    with pytest.raises(DataTransformationError):
        sim.validate_year_columns(df, mock_validation_context)


@pytest.mark.parametrize("columns",
                         (('year_start',), ('year_end',), ('year_id_start', 'year_end')),
                         ids=("missing_end", "missing_start", "typo"))
def test_validate_year_columns_missing(mock_validation_context, columns):
    df = pd.DataFrame()
    for col in columns:
        df[col] = [1, 2, 3]
    with pytest.raises(DataTransformationError, match='in columns named'):
        sim.validate_year_columns(df, mock_validation_context)


@pytest.mark.parametrize("values", [(-1, 2, 3)], ids=['integers'])
def test_validate_value_column_pass(values):
    df = pd.DataFrame({'value': values})
    sim.validate_value_column(df)


@pytest.mark.parametrize("values", [(1, 2, np.inf),
                                    (1, np.nan, 2)],
                         ids=["infinity", "missing"])
def test_validate_value_column_fail(values):
    df = pd.DataFrame({'value': values})
    with pytest.raises(DataTransformationError):
        sim.validate_value_column(df)


def test_validate_value_column_missing():
    df = pd.DataFrame({'value_column': [1, 2, 3]})
    with pytest.raises(DataTransformationError, match='in a column named'):
        sim.validate_value_column(df)


@pytest.mark.parametrize('values, ids, restriction_type, fill', [
        ((1, 1, 1, 1, 1), (1, 5), 'outer', 0.0),
        ((0, 0, 1, 1, 1), (3, 5), 'outer', 0.0),
        ((0, 1, 1, 1, 0), (2, 4), 'outer', 0.0),
        ((1, 1, 1, 0, 0), (1, 3), 'outer', 0.0),
        ((2, 2, 2, 1, 1), (1, 3), 'outer', 1.0),
], ids=('no_restr', 'left_restr', 'outer_restr', 'right_restr', 'nonzero_fill'))
def test_check_age_restrictions(mocker, mock_validation_context, values, ids, restriction_type, fill):
    entity = mocker.patch('vivarium_inputs.validation.sim.utilities.get_age_group_ids_by_restriction')
    entity.return_value = ids
    age_bins = mock_validation_context['age_bins']
    df = age_bins.filter(['age_group_start', 'age_group_end'])
    df['value'] = values
    sim.check_age_restrictions(df, entity, restriction_type, fill, mock_validation_context)


@pytest.mark.parametrize('values, ids, restriction_type, fill', [
        ((1, 1, 1, 1, 1), (1, 4), 'outer', 0.0),
        ((0, 1, 1, 1, 1), (1, 3), 'outer', 0.0),
        ((1, 1, 1, 1, 0), (1, 3), 'outer', 0.0),
        ((2, 2, 2, 2, 1), (2, 5), 'outer', 1.0),
], ids=('both_sides', 'left_side', 'right_side', 'nonzero_fill'))
def test_check_age_restrictions_fail(mocker, mock_validation_context, values, ids, restriction_type, fill):
    entity = mocker.patch('vivarium_inputs.validation.sim.utilities.get_age_group_ids_by_restriction')
    entity.return_value = ids
    age_bins = mock_validation_context['age_bins']
    df = age_bins.filter(['age_group_start', 'age_group_end'])
    df['value'] = values
    with pytest.raises(DataTransformationError):
        sim.check_age_restrictions(df, entity, restriction_type, fill, mock_validation_context)


@pytest.mark.parametrize('values, restrictions, fill', [
    ((1, 1, 1, 1), (False, False), 0.0),
    ((1, 1, 0, 0), (True, False), 0.0),
    ((0, 0, 1, 1), (False, True), 0.0),
    ((1, 1, 2, 2), (False, True), 1.0)
], ids=('None', 'male', 'female', 'nonzero_fill'))
def test_check_sex_restrictions(values, restrictions, fill):
    df = pd.DataFrame({'sex': ['Male', 'Male', 'Female', 'Female'], 'value': values})
    sim.check_sex_restrictions(df, restrictions[0], restrictions[1], fill)


@pytest.mark.parametrize('values, restrictions, fill', [
    ((1, 1, 1, 0), (True, False), 0.0),
    ((0, 1, 1, 1), (False, True), 0.0),
    ((1, 2, 2, 2), (False, True), 1.0)
], ids=('male', 'female', 'nonzero_fill'))
def test_check_sex_restrictions_fail(values, restrictions, fill):
    df = pd.DataFrame({'sex': ['Male', 'Male', 'Female', 'Female'], 'value': values})
    with pytest.raises(DataTransformationError):
        sim.check_sex_restrictions(df, restrictions[0], restrictions[1], fill)



import pytest

from pathlib import Path
import pandas as pd
import argparse

from vivarium_inputs.data_artifact import aggregation, ArtifactAggregationWarning, get_versions

from vivarium_public_health.dataset_manager import Artifact


METADATA = ['metadata.keyspace', 'metadata.locations', 'metadata.versions']


@pytest.fixture()
def single_artifact(tmpdir):
    def test_artifact(location):
        artifact_path = f'{str(tmpdir)}/test_{location}.hdf'
        artifact = Artifact(artifact_path)
        artifact.write('metadata.locations', [location.replace('_', ' ')])
        artifact.write('metadata.versions', get_versions())
        artifact.write('cause.all_causes.cause_specific_mortality',
                       pd.DataFrame({'location': location.replace('_', ' '), 'value': 0.1}, index=range(10)))
        artifact.write('risk_factor.child_stunting.exposure',
                       pd.DataFrame({'location': location.replace('_', ' '), 'value': 0.2}, index=range(10)))
        artifact.write('risk_factor.child_stunting.distribution', 'polytomous')
        return artifact
    return test_artifact


def test_unpack_arguments():
    config_path = 'path/to/test.yaml'
    output_root = 'path/to/artifact'
    locations = ['United States', 'Canada']
    args = f'--config_path {config_path} --output_root {output_root} --locations United_States Canada '.split()
    parser = argparse.ArgumentParser()
    assert aggregation.unpack_arguments(parser, args) == (Path(config_path), locations, output_root)


def test_aggregate_one_location(single_artifact, mocker):
    aggregate_mock = mocker.patch('vivarium_inputs.data_artifact.aggregation.aggregate')
    test_artifact = single_artifact('United_States')
    artifact_path = Path(test_artifact.path)
    config_path = 'path/to/test.yaml'
    output_root = artifact_path.parent.as_posix()
    locations = 'United_States'

    aggregation.main(['--config_path', config_path, '--output_root', output_root, '--locations', locations])

    renamed_artifact = artifact_path.parent / 'test.hdf'
    assert not artifact_path.is_file()
    assert renamed_artifact.is_file()
    aggregate_mock.assert_not_called()


def test_aggregate_with_missing_keys(single_artifact):

    usa_artifact = single_artifact('United_States')
    canada_artifact = single_artifact('Canada')
    kenya_artifact = single_artifact('Kenya')

    # adding a new key to only two locations
    new_key = 'new.key'
    usa_artifact.write(new_key, 'data')
    canada_artifact.write(new_key, 'data')
    artifact_path = Path(usa_artifact.path).parent.as_posix()

    with pytest.warns(ArtifactAggregationWarning) as record:
        new_artifact = aggregation.aggregate([usa_artifact, canada_artifact, kenya_artifact], f'{artifact_path}/test.hdf')

        assert len(record) == 1
        assert new_key in str(record[-1].message)
        assert 'Kenya' in str(record[-1].message)
        assert Path(new_artifact.path).is_file()
        assert set(new_artifact.keys) == set(usa_artifact.keys) == set(canada_artifact.keys)
        assert set(new_artifact.load('metadata.locations')) == {'Canada', 'United States'}

        # clean up the individual artifact used for aggregation
        assert not Path(usa_artifact.path).is_file()
        assert not Path(canada_artifact.path).is_file()
        assert not Path(kenya_artifact.path).is_file()
        # Kenya artifact should be moved to broken artifacts folder
        new_location = Path(artifact_path)/'broken_artifacts'/f'{Path(kenya_artifact.path).name}'
        assert new_location.is_file()


def test_aggregate(single_artifact):
    usa_artifact = single_artifact('United_States')
    canada_artifact = single_artifact('Canada')
    artifact_path = Path(usa_artifact.path).parent.as_posix()
    new_artifact = aggregation.aggregate([usa_artifact, canada_artifact], f'{artifact_path}/test.hdf')

    assert set(new_artifact.keys) == set(usa_artifact.keys) == set(canada_artifact.keys)
    assert set(new_artifact.load('metadata.locations')) == {'Canada', 'United States'}

    assert Path(new_artifact.path).is_file()
    assert not Path(usa_artifact.path).is_file()
    assert not Path(canada_artifact.path).is_file()


def test_disaggregate(single_artifact):
    usa_artifact = single_artifact('United_States')
    canada_artifact = single_artifact('Canada')
    kenya_artifact = single_artifact('Kenya')
    individual_artifacts = [usa_artifact, canada_artifact, kenya_artifact]
    expected_keys = usa_artifact.keys

    artifact_path = Path(usa_artifact.path).parent / 'test.hdf'
    aggregated_artifact = aggregation.aggregate(individual_artifacts, artifact_path.as_posix())
    config_name = 'test'
    output_root = Path(aggregated_artifact.path).parent

    disaggregated_artifacts_locations = aggregation.disaggregate(config_name, output_root)

    # after disaggregation, initial aggregated artifact should be deleted
    assert not Path(aggregated_artifact.path).is_file()

    for loc in disaggregated_artifacts_locations:
        l = loc.replace(' ', '_').replace("'", "-")
        new_individual_artifact_path = output_root/f'{config_name}_{l}.hdf'
        assert new_individual_artifact_path.is_file()
        new_artifact = Artifact(new_individual_artifact_path.as_posix())
        assert set(new_artifact.keys) == set(expected_keys)
        assert new_artifact.load('metadata.locations') == [loc.replace('_', ' ')]


def test_disaggregate_with_different_versions(single_artifact):
    usa_artifact = single_artifact('United_States')
    canada_artifact = single_artifact('Canada')
    kenya_artifact = single_artifact('Kenya')
    individual_artifacts = [usa_artifact, canada_artifact, kenya_artifact]

    current_versions = get_versions()
    different_versions = {k: '0.0.1' for k in current_versions}

    artifact_path = Path(usa_artifact.path).parent / 'test.hdf'
    aggregated_artifact = aggregation.aggregate(individual_artifacts, artifact_path.as_posix())

    # change the versions in aggregated_artifact
    aggregated_artifact.replace('metadata.versions', different_versions)
    assert aggregated_artifact.load('metadata.versions') == different_versions
    config_name = 'test'
    output_root = Path(aggregated_artifact.path).parent

    with pytest.warns(ArtifactAggregationWarning) as record:
        disaggregated_artifacts_locations = aggregation.disaggregate(config_name, output_root)
        assert len(record) == 1
        assert str(record[0].message) == 'Your artifact was built under the different versions and we cannot append it.' \
                                         ' It will be built from scratch'

    for loc in disaggregated_artifacts_locations:
        new_individual_artifact_path = output_root / f'{config_name}_{loc}.hdf'
        assert new_individual_artifact_path.is_file()
        new_artifact = Artifact(new_individual_artifact_path.as_posix())
        assert new_artifact.load('metadata.versions') == current_versions




import pandas as pd
import pytest

from vivarium_inputs.globals import DataAbnormalError, DataTransformationError
from vivarium_inputs.validation import shared

test_data = [(pd.DataFrame({'a': [0, 1], 'b': [2, 20]}), 1, 'lower', ['a', 'b'],
              True, 'values below', DataAbnormalError),
             (pd.DataFrame({'a': [0, 1], 'b': [2, 20]}), 0, 'lower', ['a', 'b'],
              False, 'values below or equal to', ValueError),
             (pd.DataFrame({'a': [0], 'b': [10], 'c': [100]}), 10, 'lower', ['a'],
              True, 'values below', ValueError),
             (pd.DataFrame({'a': [0, 1], 'b': [2, 20]}), 20, 'upper', ['a', 'b'],
              False, 'values above', DataAbnormalError),
             (pd.DataFrame({'a': [0, 1], 'b': [2, 20]}, index=[0, 1]), pd.Series([5, 10], index=[0, 1]), 'upper',
              ['a', 'b'], True, 'above the expected', DataTransformationError)]


@pytest.mark.parametrize('data, boundary, boundary_type, val_cols, inclusive, match, error', test_data)
def test_check_value_columns_boundary_fail_warn(data, boundary, boundary_type, val_cols, inclusive, match, error):
    with pytest.raises(error, match=match):
        shared.check_value_columns_boundary(data, boundary, boundary_type,
                                            val_cols, inclusive=inclusive, error=error)

    with pytest.warns(Warning, match=match):
        shared.check_value_columns_boundary(data, boundary, boundary_type, val_cols, inclusive=inclusive, error=None)


test_data = [(pd.DataFrame({'a': [0, 1], 'b': [2, 20]}), 0, 'lower', ['a', 'b'], True),
             (pd.DataFrame({'a': [0, 1], 'b': [2, 20]}), -1, 'lower', ['a', 'b'], False),
             (pd.DataFrame({'a': [0], 'b': [10], 'c': [100]}), 10, 'lower', ['b', 'c'], True),
             (pd.DataFrame({'a': [0, 1], 'b': [2, 20]}), 20, 'upper', ['a', 'b'], True),
             (pd.DataFrame({'a': [0, 1], 'b': [2, 20]}, index=[0, 1]), pd.Series([0, 1], index=[0, 1]), 'upper',
              ['a'], True)]


@pytest.mark.parametrize('data, boundary, boundary_type, val_cols, inclusive', test_data)
def test_check_value_columns_boundary_pass(data, boundary, boundary_type, val_cols, inclusive):
    shared.check_value_columns_boundary(data, boundary, boundary_type, val_cols,
                                        inclusive=inclusive, error=ValueError)



import pkg_resources


def get_versions():
    libraries = ['vivarium', 'vivarium_inputs', 'vivarium_public_health', 'gbd_mapping']
    return {k: pkg_resources.get_distribution(k).version for k in libraries}



"""Access to vivarium simulation input data."""
import pandas as pd

from gbd_mapping import ModelableEntity
from vivarium_inputs import core, utilities
from vivarium_inputs.globals import Population
import vivarium_inputs.validation.sim as validation


def get_measure(entity: ModelableEntity, measure: str, location: str) -> pd.DataFrame:
    """Pull GBD data for measure and entity and prep for simulation input,
    including scrubbing all GBD conventions to replace IDs with meaningful
    values or ranges and expanding over all demographic dimensions.

    Available measures:

        For entity kind 'sequela':
            incidence, prevalence, birth_prevalence, disability_weight

        For entity kind 'cause':
            incidence, prevalence, birth_prevalence, disability_weight,
            remission, cause_specific_mortality, excess_mortality

        For entity kind 'coverage_gap':
            exposure, exposure_standard_deviation, exposure_distribution_weights,
            relative_risk

        For entity kind 'risk_factor':
            exposure, exposure_standard_deviation, exposure_distribution_weights,
            relative_risk, population_attributable_fraction, mediation_factors

        For entity kind 'etiology':
            population_attributable_fraction

        For entity kind 'alternative_risk_factor':
            exposure, exposure_standard_deviation, exposure_distribution_weights

        For entity kind 'covariate':
            estimate

        For entity kind 'healthcare_entity':
            cost, utilization

        For entity kind 'health_technology':
            cost

    Parameters
    ----------
    entity
        Entity for which to pull `measure`.
    measure
        Measure for which to pull data, should be a measure available for the
        kind of entity which `entity` is.
    location
        Location for which to pull data.

    Returns
    -------
    Dataframe standardized to the format expected by `vivarium` simulations.

    """
    data = core.get_data(entity, measure, location)
    data = utilities.scrub_gbd_conventions(data, location)
    validation.validate_for_simulation(data, entity, measure, location)
    return utilities.sort_data(data)


def get_population_structure(location: str) -> pd.DataFrame:
    """Pull GBD population data for the given location and standardize to the
    expected simulation input format, including scrubbing all GBD conventions
    to replace IDs with meaningful values or ranges and expanding over all
    demographic dimensions.

    Parameters
    ----------
    location
        Location for which to pull population data.

    Returns
    -------
    Dataframe of population data for `location`, standardized to the format
    expected by `vivarium` simulations.

    """
    pop = Population()
    data = core.get_data(pop, 'structure', location)
    data = utilities.scrub_gbd_conventions(data, location)
    validation.validate_for_simulation(data, pop, 'structure', location)
    return utilities.sort_data(data)


def get_theoretical_minimum_risk_life_expectancy() -> pd.DataFrame:
    """Pull GBD theoretical minimum risk life expectancy data and standardize
    to the expected simulation input format, including binning age parameters
    as expected by simulations.

    Returns
    -------
    Dataframe of theoretical minimum risk life expectancy data, standardized
    to the format expected by `vivarium` simulations with binned age parameters.

    """
    pop = Population()
    data = core.get_data(pop, 'theoretical_minimum_risk_life_expectancy', 'Global')
    validation.validate_for_simulation(data, pop, 'theoretical_minimum_risk_life_expectancy', 'Global')
    return utilities.sort_data(data)


def get_age_bins() -> pd.DataFrame:
    """Pull GBD age bin data and standardize to the expected simulation input
    format.

    Returns
    -------
    Dataframe of age bin data, with bin start and end values as well as bin
    names.

    """
    pop = Population()
    data = core.get_data(pop, 'age_bins', 'Global')
    validation.validate_for_simulation(data, pop, 'age_bins', 'Global')
    return utilities.sort_data(data)


def get_demographic_dimensions(location: str) -> pd.DataFrame:
    """Pull the full demographic dimensions for GBD data, standardized to the
    expected simulation input format, including scrubbing all GBD conventions
    to replace IDs with with meaningful values or ranges.

    Parameters
    ----------
    location
        Location for which to pull demographic dimension data.

    Returns
    -------
    Dataframe with age and year bins from GBD, sexes, and the given location.

    """
    pop = Population()
    data = core.get_data(pop, 'demographic_dimensions', location)
    data = utilities.scrub_gbd_conventions(data, location)
    validation.validate_for_simulation(data, pop, 'demographic_dimensions', location)
    return utilities.sort_data(data)



from typing import Union
from itertools import product
import warnings

from gbd_mapping import Cause, Sequela, RiskFactor, CoverageGap, Etiology, Covariate, causes
import pandas as pd
import numpy as np

from vivarium_inputs import utilities, extract, utility_data
from vivarium_inputs.globals import (InvalidQueryError, DEMOGRAPHIC_COLUMNS, MEASURES, SEXES, DRAW_COLUMNS,
                                     Population, DataDoesNotExistError)
from vivarium_inputs.mapping_extension import AlternativeRiskFactor, HealthcareEntity, HealthTechnology


def get_data(entity, measure: str, location: str):
    measure_handlers = {
        # Cause-like measures
        'incidence': (get_incidence, ('cause', 'sequela')),
        'prevalence': (get_prevalence, ('cause', 'sequela')),
        'birth_prevalence': (get_birth_prevalence, ('cause', 'sequela')),
        'disability_weight': (get_disability_weight, ('cause', 'sequela')),
        'remission': (get_remission, ('cause',)),
        'cause_specific_mortality': (get_cause_specific_mortality, ('cause',)),
        'excess_mortality': (get_excess_mortality, ('cause',)),
        'case_fatality': (get_case_fatality, ('cause',)),
        # Risk-like measures
        'exposure': (get_exposure, ('risk_factor', 'coverage_gap', 'alternative_risk_factor',)),
        'exposure_standard_deviation': (get_exposure_standard_deviation, ('risk_factor', 'alternative_risk_factor')),
        'exposure_distribution_weights': (get_exposure_distribution_weights, ('risk_factor', 'alternative_risk_factor')),
        'relative_risk': (get_relative_risk, ('risk_factor', 'coverage_gap')),
        'population_attributable_fraction': (get_population_attributable_fraction, ('risk_factor', 'etiology')),
        'mediation_factors': (get_mediation_factors, ('risk_factor',)),
        # Covariate measures
        'estimate': (get_estimate, ('covariate',)),
        # Health system measures
        'cost': (get_cost, ('healthcare_entity', 'health_technology')),
        'utilization': (get_utilization, ('healthcare_entity',)),
        # Population measures
        'structure': (get_structure, ('population',)),
        'theoretical_minimum_risk_life_expectancy': (get_theoretical_minimum_risk_life_expectancy, ('population',)),
        'age_bins': (get_age_bins, ('population',)),
        'demographic_dimensions': (get_demographic_dimensions, ('population',))

    }

    if measure not in measure_handlers:
        raise InvalidQueryError(f'No functions available to pull data for measure {measure}.')

    handler, entity_types = measure_handlers[measure]

    if entity.kind not in entity_types:
        raise InvalidQueryError(f'{measure.capitalize()} not available for {entity.kind}.')

    location_id = utility_data.get_location_id(location)
    data = handler(entity, location_id)
    return data


def get_incidence(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'incidence', location_id)
    if entity.kind == 'cause':
        restrictions_entity = entity
    else:  # sequela
        cause = [c for c in causes if c.sequelae and entity in c.sequelae][0]
        restrictions_entity = cause

    data = utilities.filter_data_by_restrictions(data, restrictions_entity,
                                                 'yld', utility_data.get_age_group_ids())
    data = utilities.normalize(data, fill_value=0)
    data = utilities.reshape(data).set_index(DEMOGRAPHIC_COLUMNS + ['draw'])
    prevalence = get_prevalence(entity, location_id).set_index(DEMOGRAPHIC_COLUMNS + ['draw'])
    # Convert from "True incidence" to the incidence rate among susceptibles
    data /= 1 - prevalence
    return data.fillna(0).reset_index()


def get_prevalence(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'prevalence', location_id)
    if entity.kind == 'cause':
        restrictions_entity = entity
    else:  # sequela
        cause = [c for c in causes if c.sequelae and entity in c.sequelae][0]
        restrictions_entity = cause

    data = utilities.filter_data_by_restrictions(data, restrictions_entity,
                                                 'yld', utility_data.get_age_group_ids())
    data = utilities.normalize(data, fill_value=0)
    data = utilities.reshape(data)
    return data


def get_birth_prevalence(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'birth_prevalence', location_id)
    data = data.drop('age_group_id', 'columns')
    data = utilities.normalize(data, fill_value=0)
    data = utilities.reshape(data, to_keep=('year_id', 'sex_id', 'location_id'))
    return data


def get_disability_weight(entity: Union[Cause, Sequela], location_id: int) -> pd.DataFrame:
    if entity.kind == 'cause':
        data = get_demographic_dimensions(Population(), location_id, draws=True)
        data['value'] = 0.0
        data = data.set_index(DEMOGRAPHIC_COLUMNS + ['draw'])
        if entity.sequelae:
            for sequela in entity.sequelae:
                try:
                    prevalence = get_prevalence(sequela, location_id).set_index(DEMOGRAPHIC_COLUMNS + ['draw'])
                except DataDoesNotExistError:
                    # sequela prevalence does not exist so no point continuing with this sequela
                    continue
                disability = get_disability_weight(sequela, location_id)
                disability['location_id'] = location_id
                disability = disability.set_index(DEMOGRAPHIC_COLUMNS + ['draw'])
                data += prevalence * disability
        data = data.reset_index()
    else:  # entity.kind == 'sequela'
        if not entity.healthstate.disability_weight_exists:
            data = get_demographic_dimensions(Population(), location_id, draws=True)
            data['value'] = 0.0
        else:
            data = extract.extract_data(entity, 'disability_weight', location_id)
            data = utilities.normalize(data)
            cause = [c for c in causes if c.sequelae and entity in c.sequelae][0]
            data = utilities.clear_disability_weight_outside_restrictions(data, cause, 0.0,
                                                                          utility_data.get_age_group_ids())
            data = utilities.reshape(data)

    return data


def get_remission(entity: Cause, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'remission', location_id)

    data = utilities.filter_data_by_restrictions(data, entity,
                                                 'yld', utility_data.get_age_group_ids())
    data = utilities.normalize(data, fill_value=0)
    data = utilities.reshape(data)
    return data


def get_cause_specific_mortality(entity: Cause, location_id: int) -> pd.DataFrame:
    deaths = _get_deaths(entity, location_id)
    pop = get_structure(Population(), location_id)
    data = deaths.merge(pop, on=DEMOGRAPHIC_COLUMNS)
    data['value'] = data['value_x'] / data['value_y']
    return data.drop(['value_x', 'value_y'], 'columns')


def get_excess_mortality(entity: Cause, location_id: int) -> pd.DataFrame:
    csmr = get_cause_specific_mortality(entity, location_id).set_index(DEMOGRAPHIC_COLUMNS + ['draw'])
    prevalence = get_prevalence(entity, location_id).set_index(DEMOGRAPHIC_COLUMNS + ['draw'])
    data = (csmr / prevalence).fillna(0)
    data = data.replace([np.inf, -np.inf], 0)
    return data.reset_index()


def get_case_fatality(entity: Cause, location_id: int):
    raise NotImplementedError()


def _get_deaths(entity: Cause, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'deaths', location_id)
    data = utilities.filter_data_by_restrictions(data, entity,
                                                 'yll', utility_data.get_age_group_ids())
    data = utilities.normalize(data, fill_value=0)
    data = utilities.reshape(data)
    return data


def get_exposure(entity: Union[RiskFactor, AlternativeRiskFactor, CoverageGap], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'exposure', location_id)
    data = data.drop('modelable_entity_id', 'columns')

    if entity.kind in ['risk_factor', 'alternative_risk_factor']:
        data = utilities.filter_data_by_restrictions(data, entity,
                                                     'outer', utility_data.get_age_group_ids())

    if entity.distribution in ['dichotomous', 'ordered_polytomous', 'unordered_polytomous']:
        tmrel_cat = sorted(list(entity.categories.to_dict()), key=lambda x: int(x[3:]))[-1]
        exposed = data[data.parameter != tmrel_cat]
        unexposed = data[data.parameter == tmrel_cat]

        #  FIXME: We fill 1 as exposure of tmrel category, which is not correct.
        data = pd.concat([utilities.normalize(exposed, fill_value=0), utilities.normalize(unexposed, fill_value=1)],
                         ignore_index=True)

        # normalize so all categories sum to 1
        cols = list(set(data.columns).difference(DRAW_COLUMNS + ['parameter']))
        sums = data.groupby(cols)[DRAW_COLUMNS].sum()
        data = (data.groupby('parameter')
                .apply(lambda df: df.set_index(cols).loc[:, DRAW_COLUMNS].divide(sums))
                .reset_index())
    else:
        data = utilities.normalize(data, fill_value=0)
    data = utilities.reshape(data, to_keep=DEMOGRAPHIC_COLUMNS + ['parameter'])
    return data


def get_exposure_standard_deviation(entity: Union[RiskFactor, AlternativeRiskFactor], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'exposure_standard_deviation', location_id)
    data = data.drop('modelable_entity_id', 'columns')

    exposure = extract.extract_data(entity, 'exposure', location_id)
    valid_age_groups = utilities.get_exposure_and_restriction_ages(exposure, entity)
    data = data[data.age_group_id.isin(valid_age_groups)]

    data = utilities.normalize(data, fill_value=0)
    data = utilities.reshape(data)
    return data


def get_exposure_distribution_weights(entity: Union[RiskFactor, AlternativeRiskFactor], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'exposure_distribution_weights', location_id)

    exposure = extract.extract_data(entity, 'exposure', location_id)
    valid_ages = utilities.get_exposure_and_restriction_ages(exposure, entity)

    data.drop('age_group_id', axis=1, inplace=True)
    df = []
    for age_id in valid_ages:
        copied = data.copy()
        copied['age_group_id'] = age_id
        df.append(copied)
    data = pd.concat(df)
    distribution_cols = ['exp', 'gamma', 'invgamma', 'llogis', 'gumbel', 'invweibull', 'weibull',
                         'lnorm', 'norm', 'glnorm', 'betasr', 'mgamma', 'mgumbel']

    data = utilities.normalize(data, fill_value=0, cols_to_fill=distribution_cols)
    id_cols = ['rei_id', 'location_id', 'sex_id', 'age_group_id', 'measure', 'year_id']
    data = pd.melt(data, id_vars=id_cols, value_vars=distribution_cols, var_name='parameter')
    return data


def filter_relative_risk_to_cause_restrictions(data: pd.DataFrame) -> pd.DataFrame:
    """ It applies age restrictions according to affected causes
    and affected measures. If affected measure is incidence_rate,
    it applies the yld_age_restrictions. If affected measure is
    excess_mortality, it applies the yll_age_restrictions to filter
    the relative_risk data"""

    causes_map = {c.name: c for c in causes}
    temp = []
    affected_entities = set(data.affected_entity)
    affected_measures = set(data.affected_measure)
    for cause, measure in product(affected_entities, affected_measures):
        df = data[(data.affected_entity == cause) & (data.affected_measure)]
        cause = causes_map[cause]
        if measure == 'excess_mortality':
            start, end = utilities.get_age_group_ids_by_restriction(cause, 'yll')
        else:  # incidence_rate
            start, end = utilities.get_age_group_ids_by_restriction(cause, 'yld')
        temp.append(df[df.age_group_id.isin(range(start, end + 1))])
    data = pd.concat(temp)
    return data


def get_relative_risk(entity: Union[RiskFactor, CoverageGap], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'relative_risk', location_id)

    if entity.kind == 'risk_factor':
        # FIXME: we don't currently support yll-only causes so I'm dropping them because the data in some cases is
        #  very messed up, with mort = morb = 1 (e.g., aortic aneurysm in the RR data for high systolic bp) -
        #  2/8/19 K.W.
        yll_only_causes = set([c.gbd_id for c in causes if c.restrictions.yll_only])
        data = data[~data.cause_id.isin(yll_only_causes)]

        data = utilities.convert_affected_entity(data, 'cause_id')
        morbidity = data.morbidity == 1
        mortality = data.mortality == 1
        data.loc[morbidity & mortality, 'affected_measure'] = 'incidence_rate'
        data.loc[morbidity & ~mortality, 'affected_measure'] = 'incidence_rate'
        data.loc[~morbidity & mortality, 'affected_measure'] = 'excess_mortality'
        data = filter_relative_risk_to_cause_restrictions(data)
    else:  # coverage_gap
        data = utilities.convert_affected_entity(data, 'rei_id')
        data['affected_measure'] = 'exposure_parameters'

    result = []
    for affected_entity in data.affected_entity.unique():
        df = data[data.affected_entity == affected_entity]
        df = df.groupby('parameter').apply(lambda d: utilities.normalize(d, fill_value=1))
        result.append(df)
    data = pd.concat(result)
    data = utilities.reshape(data, to_keep=DEMOGRAPHIC_COLUMNS + ['affected_entity', 'affected_measure', 'parameter'])

    if entity.distribution in ['dichotomous', 'ordered_polytomous', 'unordered_polytomous']:
        tmrel_cat = sorted(list(entity.categories.to_dict()), key=lambda x: int(x[3:]))[-1]
        if np.allclose(data.loc[data.parameter == tmrel_cat, 'value'], 1.0):
            data.loc[data.parameter == tmrel_cat, 'value'] = 1.0

    return data


def filter_by_relative_risk(df: pd.DataFrame, relative_risk: pd.DataFrame) -> pd.DataFrame:
    c_id = df.cause_id.unique()[0]
    rr = relative_risk[relative_risk.cause_id == c_id]
    #  We presume all attributable mortality moves through incidence.
    if set(rr.mortality) == {1} and set(rr.morbidity) == {1}:
        df = df[df.measure_id == MEASURES['YLDs']]
    return df


def get_population_attributable_fraction(entity: Union[RiskFactor, Etiology], location_id: int) -> pd.DataFrame:
    causes_map = {c.gbd_id: c for c in causes}
    if entity.kind == 'risk_factor':
        data = extract.extract_data(entity, 'population_attributable_fraction', location_id)
        relative_risk = extract.extract_data(entity, 'relative_risk', location_id)

        # FIXME: we don't currently support yll-only causes so I'm dropping them because the data in some cases is
        #  very messed up, with mort = morb = 1 (e.g., aortic aneurysm in the RR data for high systolic bp) -
        #  2/8/19 K.W.
        yll_only_causes = set([c.gbd_id for c in causes if c.restrictions.yll_only])
        data = data[~data.cause_id.isin(yll_only_causes)]
        relative_risk = relative_risk[~relative_risk.cause_id.isin(yll_only_causes)]

        data = data.groupby('cause_id', as_index=False).apply(filter_by_relative_risk, relative_risk).reset_index(drop=True)

        temp = []
        # We filter paf age groups by cause level restrictions.
        for (c_id, measure), df in data.groupby(['cause_id', 'measure_id']):
            cause = causes_map[c_id]
            measure = 'yll' if measure == MEASURES['YLLs'] else 'yld'
            df = utilities.filter_data_by_restrictions(df, cause, measure, utility_data.get_age_group_ids())
            temp.append(df)
        data = pd.concat(temp, ignore_index=True)

    else:  # etiology
        data = extract.extract_data(entity, 'etiology_population_attributable_fraction', location_id)
        cause = [c for c in causes if entity in c.etiologies][0]
        data = utilities.filter_data_by_restrictions(data, cause, 'inner', utility_data.get_age_group_ids())
        if np.any(data[DRAW_COLUMNS] < 0):
            warnings.warn(f"{entity.name.capitalize()} has negative values for paf. These will be replaced with 0.")
            other_cols = [c for c in data.columns if c not in DRAW_COLUMNS]
            data.set_index(other_cols, inplace=True)
            data = data.where(data[DRAW_COLUMNS] > 0, 0).reset_index()

    data = utilities.convert_affected_entity(data, 'cause_id')
    data.loc[data['measure_id'] == MEASURES['YLLs'], 'affected_measure'] = 'excess_mortality'
    data.loc[data['measure_id'] == MEASURES['YLDs'], 'affected_measure'] = 'incidence_rate'
    data = data.groupby(['affected_entity', 'affected_measure']).apply(lambda df: utilities.normalize(df, fill_value=0))
    data = utilities.reshape(data, to_keep=DEMOGRAPHIC_COLUMNS + ['affected_entity', 'affected_measure'])
    return data


def get_mediation_factors(entity, location_id):
    raise NotImplementedError()


def get_estimate(entity: Covariate, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'estimate', location_id)

    key_columns = ['location_id', 'year_id']
    if entity.by_age:
        key_columns.append('age_group_id')
    if entity.by_sex:
        key_columns.append('sex_id')

    data = pd.melt(data, id_vars=key_columns,
                   value_vars=['mean_value', 'upper_value', 'lower_value'], var_name='parameter')
    data = utilities.normalize(data)
    return data


def get_cost(entity: Union[HealthcareEntity, HealthTechnology], location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'cost', location_id)
    data = utilities.normalize(data, fill_value=0)
    data = utilities.reshape(data, to_keep=DEMOGRAPHIC_COLUMNS + [entity.kind])
    return data


def get_utilization(entity: HealthcareEntity, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'utilization', location_id)
    data = utilities.normalize(data, fill_value=0)
    data = utilities.reshape(data)
    return data


def get_structure(entity: Population, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'structure', location_id)
    data = data.drop('run_id', 'columns').rename(columns={'population': 'value'})
    data = utilities.normalize(data)
    return data


def get_theoretical_minimum_risk_life_expectancy(entity: Population, location_id: int) -> pd.DataFrame:
    data = extract.extract_data(entity, 'theoretical_minimum_risk_life_expectancy', location_id)
    data = data.rename(columns={'age': 'age_group_start', 'life_expectancy': 'value'})
    data['age_group_end'] = data.age_group_start.shift(-1).fillna(125.)
    return data


def get_age_bins(entity: Population, location_id: int) -> pd.DataFrame:
    age_bins = utility_data.get_age_bins()[['age_group_name', 'age_group_start', 'age_group_end']]
    return age_bins


def get_demographic_dimensions(entity: Population, location_id: int, draws: bool = False) -> pd.DataFrame:
    ages = utility_data.get_age_group_ids()
    estimation_years = utility_data.get_estimation_years()
    years = range(min(estimation_years), max(estimation_years) + 1)
    sexes = [SEXES['Male'], SEXES['Female']]
    location = [location_id]
    values = [location, sexes, ages, years]
    names = ['location_id', 'sex_id', 'age_group_id', 'year_id']
    if draws:
        values.append(range(1000))
        names.append('draw')

    demographic_dimensions = (pd.MultiIndex
                              .from_product(values, names=names)
                              .to_frame(index=False))
    demographic_dimensions = utilities.normalize(demographic_dimensions)
    return demographic_dimensions



import numpy as np
import pandas as pd
import pytest

from vivarium_inputs.validation import raw
from vivarium_inputs.globals import DataAbnormalError, DataDoesNotExistError, SEXES

LOCATION_ID = 100
ESTIMATION_YEARS = tuple(list(range(1990, 2015, 5)) + [2017])
AGE_GROUP_IDS = tuple(range(1, 6))
PARENT_LOCATIONS = (1, 5, 10)

@pytest.fixture
def mock_validation_context():
    context = raw.RawValidationContext(
        location_id=LOCATION_ID,
        estimation_years=list(ESTIMATION_YEARS),
        age_group_ids=list(AGE_GROUP_IDS),
        sexes=SEXES,
        parent_locations=list(PARENT_LOCATIONS),
    )
    return context


@pytest.fixture
def measures_mock(mocker):
    return mocker.patch('vivarium_inputs.validation.raw.MEASURES', {'A': 1, 'B': 2})


@pytest.fixture
def metrics_mock(mocker):
    return mocker.patch('vivarium_inputs.validation.raw.METRICS', {'A': 1, 'B': 2})


@pytest.mark.parametrize('mort, morb, yld_only, yll_only, match',
                         [([0, 1, 1], [0, 1, 1], True, False, 'set to 0'),
                          ([1, 1, 0], [1, 1, 1], True, False, 'only one'),
                          ([1, 1, 1], [0, 0, 0], False, False, 'not restricted to yll_only'),
                          ([0, 0, 0], [1, 1, 1], False, False, 'not restricted to yld_only'),
                          ([1, 0, 0], [0, 1, 1], True, False, 'is restricted to yld_only'),
                          ([1, 0, 2], [0, 1, 1], False, False, 'outside the expected')])
def test_check_mort_morb_flags_fail(mort, morb, yld_only, yll_only, match):
    data = pd.DataFrame({'mortality': mort, 'morbidity': morb})
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_mort_morb_flags(data, yld_only, yll_only)


@pytest.mark.parametrize('mort, morb, yld_only, yll_only',
                         [([1, 1, 1], [1, 1, 1], True, False),
                          ([1, 1, 1], [1, 1, 1], False, False),
                          ([0, 0, 0], [1, 1, 1], True, False,),
                          ([1, 1, 1], [0, 0, 0], False, True,),
                          ([1, 0, 1], [0, 1, 0], False, False)])
def test_check_mort_morb_flags_pass(mort, morb, yld_only, yll_only):
    data = pd.DataFrame({'mortality': mort, 'morbidity': morb})
    raw.check_mort_morb_flags(data, yld_only, yll_only)


@pytest.mark.parametrize('years, bin_type', [(range(1950, 2020), 'annual'),
                                             (range(1990, 2018), 'annual'),
                                             (list(range(1990, 2015, 5)) + [2017], 'binned')],
                         ids=['annual with extra', 'annual', 'binned'])
def test_check_years_pass(mock_validation_context, years, bin_type):
    df = pd.DataFrame({'year_id': years})
    raw.check_years(df, mock_validation_context, bin_type)


@pytest.mark.parametrize('years, bin_type, match', [([1990, 1992], 'annual', 'missing'),
                                                    ([1990, 1995], 'binned', 'missing'),
                                                    (list(range(1990, 2015, 5)) + [2017, 2020], 'binned', 'extra')],
                         ids=['annual with gap', 'binned with gap', 'binned with extra'])
def test_check_years_fail(mock_validation_context, years, bin_type, match):
    df = pd.DataFrame({'year_id': years})
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_years(df, mock_validation_context, bin_type)


@pytest.mark.parametrize('location_id', list(range(2, 10)))
def test_check_location_pass(mock_validation_context, location_id):
    mock_validation_context['location_id'] = location_id
    df = pd.DataFrame({'location_id': [location_id] * 5})
    raw.check_location(df, mock_validation_context)


@pytest.mark.parametrize('location_id', PARENT_LOCATIONS)
def test_check_location_parent_pass(mock_validation_context, location_id):
    df = pd.DataFrame({'location_id': [location_id]})
    raw.check_location(df, mock_validation_context)


@pytest.mark.parametrize('location_ids, match', [([1, 2], 'multiple'),
                                                 ([2, 2], 'actually has location id')],
                         ids=['multiple locations', 'wrong location'])
def test_check_location_fail(mock_validation_context, location_ids, match):
    df = pd.DataFrame({'location_id': location_ids})
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_location(df, mock_validation_context)


@pytest.mark.parametrize('columns', [['a', 'b'], ['c', 'd', 'e'], ['a'], []])
def test_check_columns_pass(columns):
    raw.check_columns(columns, columns)


@pytest.mark.parametrize('columns', [['a', 'b'], ['c', 'd', 'e'], ['a'], []])
def test_check_columns_extra_fail(columns):
    with pytest.raises(DataAbnormalError, match='extra columns'):
        raw.check_columns(columns, columns+['extra'])


@pytest.mark.parametrize('columns', [['a', 'b'], ['c', 'd', 'e'], ['a']])
def test_check_columns_missing_fail(columns):
    with pytest.raises(DataAbnormalError, match='missing columns'):
        raw.check_columns(columns, columns[:-1])


@pytest.mark.parametrize('data', [pd.DataFrame({'a': [1], 'b': [np.nan], 'c': [0]}),
                                  pd.DataFrame({'a': [np.nan], 'b': [np.nan]}),
                                  pd.DataFrame({'a': [np.inf], 'b': [1], 'c': [0]}),
                                  pd.DataFrame({'a': [np.inf], 'b': [np.inf]}),
                                  pd.DataFrame(columns=['a', 'b'])],
                         ids=['Single NaN', 'All NaN', 'Single Inf', 'All Inf', 'Empty'])
def test_check_data_exist_always_fail(data):
    with pytest.raises(DataDoesNotExistError):
        raw.check_data_exist(data, value_columns=data.columns, zeros_missing=False)

    assert raw.check_data_exist(data, value_columns=data.columns, zeros_missing=False, error=False) is False

    with pytest.raises(DataDoesNotExistError):
        raw.check_data_exist(data, value_columns=data.columns, zeros_missing=True)

    assert raw.check_data_exist(data, value_columns=data.columns, zeros_missing=True, error=False) is False


def test_check_data_exist_fails_only_on_missing_zeros():
    df = pd.DataFrame({'a': [0, 0], 'b': [0, 0]})

    raw.check_data_exist(df, value_columns=df.columns, zeros_missing=False)

    with pytest.raises(DataDoesNotExistError):
        raw.check_data_exist(df, value_columns=df.columns, zeros_missing=True)


def test_check_data_exist_value_columns():
    df = pd.DataFrame({'a': [0, 0], 'b': [0, 0], 'c': [1, 1]})

    assert raw.check_data_exist(df, value_columns=['a', 'b', 'c'], zeros_missing=True, error=False)

    assert raw.check_data_exist(df, value_columns=['a', 'b'], zeros_missing=True, error=False) is False


@pytest.mark.parametrize('data', [pd.DataFrame({'a': [1], 'b': [1], 'c': [0]}),
                                  pd.DataFrame({'a': [0], 'b': [1]}),
                                  pd.DataFrame({'a': [0.0001], 'b': [0], 'c': [0]}),
                                  pd.DataFrame({'a': [1000], 'b': [-1000]})])
def test_check_data_exist_pass(data):
    assert raw.check_data_exist(data, zeros_missing=True, value_columns=data.columns, error=False)


@pytest.mark.parametrize('test_age_ids, start, end, match', [([1, 2, 3, 100], None, None, 'invalid'),
                                                             ([1, 3, 4, 5], None, None, 'non-contiguous'),
                                                             ([1, 2, 3, 100], 1, 3, 'invalid'),
                                                             ([1, 2, 3, 5], 1, 3, 'non-contiguous')])
def test_check_age_group_ids_fail(mock_validation_context, test_age_ids, start, end, match):
    df = pd.DataFrame({'age_group_id': test_age_ids})
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_age_group_ids(df, mock_validation_context, start, end)


@pytest.mark.parametrize('test_age_ids, start, end, match',
                         [([2, 3], 2, 4, 'contain all age groups in restriction range'),
                          ([2, 3, 4], 2, 3, 'additional age groups'),
                          ([1, 2, 3, 4, 5], 1, 3, 'additional age groups'),
                          ([1, 2, 3], 1, 5, 'contain all age groups in restriction range')])
def test_check_age_group_ids_warn(mock_validation_context, test_age_ids, start, end, match):
    df = pd.DataFrame({'age_group_id': test_age_ids})
    with pytest.warns(Warning, match=match):
        raw.check_age_group_ids(df, mock_validation_context, start, end)


@pytest.mark.parametrize('test_age_ids, start, end',
                         [([2, 3, 4], 2, 4),
                          ([2, 3, 4], None, None),
                          ([1, 2, 3, 4, 5], 1, 5)])
def test_check_age_group_ids_pass(mock_validation_context, test_age_ids, start, end, recwarn):
    df = pd.DataFrame({'age_group_id': test_age_ids})
    raw.check_age_group_ids(df, mock_validation_context, start, end)

    assert len(recwarn) == 0, 'An unexpected warning was raised.'


@pytest.mark.parametrize('sex_ids, male, female, both', [([1, 2, 12], True, False, False),
                                                         ([1, 2, 3, 3, 0], True, True, True),
                                                         ([5], False, False, False)])
def test_check_sex_ids_fail(mock_validation_context, sex_ids, male, female, both):
    df = pd.DataFrame({'sex_id': sex_ids})
    with pytest.raises(DataAbnormalError):
        raw.check_sex_ids(df, mock_validation_context, male, female, both)


@pytest.mark.parametrize('sex_ids, male, female, both, match', [([1, 2, 1], True, True, True, ['missing']),
                                                                ([1, 2, 3, 2], True, True, False, ['extra']),
                                                                ([1], False, True, False, ['extra', 'missing'])])
def test_check_sex_ids_warn(mock_validation_context, sex_ids, male, female, both, match):
    df = pd.DataFrame({'sex_id': sex_ids})
    with pytest.warns(None) as record:
        raw.check_sex_ids(df, mock_validation_context, male, female, both)

    assert len(record) == len(match), 'The expected number of warnings were not raised.'
    for i, m in enumerate(match):
        assert m in str(record[i].message), f'The expected warning message was not raised for warning {i+1}.'


@pytest.mark.parametrize('sex_ids, male, female, both', [([1, 1, 1], True, False, False),
                                                         ([1, 2, 2, 2], True, True, False),
                                                         ([1, 2, 3, 3, 2, 1], True, True, True),
                                                         ([], False, False, False)])
def test_check_sex_ids_pass(mock_validation_context, sex_ids, male, female, both, recwarn):
    df = pd.DataFrame({'sex_id': sex_ids})
    raw.check_sex_ids(df, mock_validation_context, male, female, both)

    assert len(recwarn) == 0, 'An unexpected warning was raised.'


test_data = [
    (pd.DataFrame({'age_group_id': [1, 2, 3], 'a': 1, 'b': 0}), 1, 4, ['a', 'b'], 'missing'),
    (pd.DataFrame({'age_group_id': [1, 2], 'a': [0, 0], 'b': [0, 0]}), 1, 2, ['a', 'b'], 'missing for all age groups')
]


@pytest.mark.parametrize('data, start, end, val_cols, match', test_data)
def test_check_age_restrictions_fail(mock_validation_context, data, start, end, val_cols, match):
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_age_restrictions(data, mock_validation_context, start, end, value_columns=val_cols)


test_data = [(pd.DataFrame({'age_group_id': [1, 2, 3, 4, 5], 'a': 1, 'b': 0}), 1, 4, ['a', 'b']),
             (pd.DataFrame({'age_group_id': [1, 2, 3], 'a': [1, 1, 1], 'b': [2, 3, 0]}), 1, 2, ['a'])]


@pytest.mark.parametrize('data, start, end, val_cols', test_data)
def test_check_age_restrictions_warn(mock_validation_context, data, start, end, val_cols):
    with pytest.warns(Warning, match='also included'):
        raw.check_age_restrictions(data, mock_validation_context, start, end, value_columns=val_cols)


test_data = [(pd.DataFrame({'age_group_id': [1, 2, 3], 'a': 1, 'b': 0}), 1, 3, ['a', 'b']),
             (pd.DataFrame({'age_group_id': [1, 2], 'a': [1, 0], 'b': [1, 0.1]}), 1, 2, ['a', 'b']),
             (pd.DataFrame({'age_group_id': [1, 2, 3], 'a': [1, 1, 0], 'b': [2, 3, 0]}), 1, 3, ['a'])]


@pytest.mark.parametrize('data, start, end, val_cols', test_data)
def test_check_age_restrictions_pass(mock_validation_context, data, start, end, val_cols):
    raw.check_age_restrictions(data, mock_validation_context, start, end, value_columns=val_cols)


test_data = [(pd.DataFrame({'sex_id': [1, 1], 'a': 0, 'b': 0, 'c': 1}), True, False,
              ['a', 'b'], 'missing data values for males'),
             (pd.DataFrame({'sex_id': [2, 2, 2, 3], 'a': 0, 'b': 0}), False, True,
              ['a', 'b'], 'missing data values for females'),
             (pd.DataFrame({'sex_id': [3], 'a': [0], 'b': [0]}), False, False,
             ['a', 'b'], 'values for both'),
             (pd.DataFrame({'sex_id': [1, 2], 'a': [0, 1], 'b': [0, 1]}), False, False,
              ['a', 'b'], 'values for both')]


@pytest.mark.parametrize('data, male, female, val_cols, match', test_data)
def test_check_sex_restrictions_fail(mock_validation_context, data, male, female, val_cols, match):
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_sex_restrictions(data, mock_validation_context, male, female, value_columns=val_cols)


test_data = [(pd.DataFrame({'sex_id': [1, 2], 'a': 1, 'b': 0}), True, False,
              ['a', 'b'], 'non-male sex ids'),
             (pd.DataFrame({'sex_id': [1, 2], 'a': 1, 'b': 0}), False, True,
              ['a', 'b'], 'non-female sex ids')]


@pytest.mark.parametrize('data, male, female, val_cols, match', test_data)
def test_check_sex_restrictions_warn(mock_validation_context, data, male, female, val_cols, match):
    with pytest.warns(Warning, match=match):
        raw.check_sex_restrictions(data, mock_validation_context, male, female, value_columns=val_cols)


test_data = [(pd.DataFrame({'sex_id': [1, 1], 'a': 1, 'b': 0, 'c': 0}), True, False, ['a', 'b']),
             (pd.DataFrame({'sex_id': [2, 2], 'a': 1, 'b': 0}), False, True, ['a', 'b']),
             (pd.DataFrame({'sex_id': [3, 3, 3], 'a': 1, 'b': 0}), False, False, ['a']),
             (pd.DataFrame({'sex_id': [1, 2], 'a': [1, 0], 'b': [0, 1]}), False, False, ['a', 'b']),
             (pd.DataFrame({'sex_id': [3], 'a': [0], 'b': [1]}), False, False, ['a', 'b'])]


@pytest.mark.parametrize('data, male, female, val_cols', test_data)
def test_check_sex_restrictions_pass(mock_validation_context, data, male, female, val_cols):
    raw.check_sex_restrictions(data, mock_validation_context, male, female, value_columns=val_cols)


@pytest.mark.parametrize('m_ids, expected, match', [([1, 1, 1, 12], ['A'], 'multiple'),
                                                    ([3, 3, 3], ['A', 'B'], 'not in the expected')])
def test_check_measure_id_fail(m_ids, expected, match, measures_mock):
    df = pd.DataFrame({'measure_id': m_ids})
    with pytest.raises(DataAbnormalError, match=match):
        raw.check_measure_id(df, expected)


@pytest.mark.parametrize('m_ids, expected', [([1], ['A']),
                                             ([2, 2, 2, 2], ['A', 'B'])])
def test_check_measure_id_pass(m_ids, expected, measures_mock):
    df = pd.DataFrame({'measure_id': m_ids})
    raw.check_measure_id(df, expected)


@pytest.mark.parametrize('m_ids, expected', [([1, 1, 1, 12], 'A'),
                                             ([3, 3, 3], 'A')])
def test_check_metric_id_fail(m_ids, expected, metrics_mock):
    df = pd.DataFrame({'metric_id': m_ids})
    with pytest.raises(DataAbnormalError):
        raw.check_metric_id(df, expected)


@pytest.mark.parametrize('m_ids, expected', [([1], 'A'),
                                             ([2, 2, 2, 2], 'B')])
def test_check_metric_id_pass(m_ids, expected, metrics_mock):
    df = pd.DataFrame({'metric_id': m_ids})
    raw.check_metric_id(df, expected)



import pytest

import pandas as pd

from vivarium_inputs import utilities


@pytest.mark.parametrize("sex_ids", [
    (1, 1, 1, 2, 2, 2),
    (1, 1, 2, 2, 3, 3),
    (1, 1, 1),
    (2, 2, 2),
    (3, 3, 3)
], ids=['male_female', 'male_female_both', 'male', 'female', 'both'])
def test_normalize_sex(sex_ids):
    df = pd.DataFrame({'sex_id': sex_ids, 'value': [1] * len(sex_ids)})
    normalized = utilities.normalize_sex(df, fill_value=0.0, cols_to_fill=['value'])
    assert {1, 2} == set(normalized.sex_id)


def test_normalize_sex_copy_3():
    values = [1, 2, 3, 4]
    df = pd.DataFrame({'sex_id': [3] * len(values), 'value': values})
    normalized = utilities.normalize_sex(df, fill_value=0.0, cols_to_fill=['value'])
    assert (normalized.loc[normalized.sex_id == 1, 'value'] == values).all()
    assert (normalized.loc[normalized.sex_id == 2, 'value'] == values).all()


def test_normalize_sex_fill_value():
    values = [1, 2, 3, 4]
    fill = 0.0
    for sex in [1, 2]:
        missing_sex = 1 if sex == 2 else 2
        df = pd.DataFrame({'sex_id': [sex] * len(values), 'value': values})
        normalized = utilities.normalize_sex(df, fill_value=fill, cols_to_fill=['value'])
        assert (normalized.loc[normalized.sex_id == sex, 'value'] == values).all()
        assert (normalized.loc[normalized.sex_id == missing_sex, 'value'] == [fill] * len(values)).all()


def test_normalize_sex_no_sex_id():
    df = pd.DataFrame({"ColumnA": [1, 2, 3], "ColumnB": [1, 2, 3]})
    normalized = utilities.normalize_sex(df, fill_value=0.0, cols_to_fill=['value'])
    pd.testing.assert_frame_equal(df, normalized)




from .health_technology_template import HealthTechnologies, HealthTechnology


health_technologies = HealthTechnologies(
    hypertension_drugs=HealthTechnology(
        name='hypertension_drugs',
        kind='health_technology',
        gbd_id=None,
    )
)
